function [fitness, pareto] = moga_fitness (x, z)
  debugprint(1,'moga_fitness: population is', [x; z])
  %% get sizes...
  [d n] = size (x);                   % dimension and number of points
  m = size (z, 1);                    % number of objective functions

  %% the first step is to identify the pareto set, i.e. the set of
  %% points within the population that best approximate the pareto
  %% front. a point in this set is one which is not dominated by any
  %% other. a point x1 dominates x2 if z1 <= z2 for all the objective
  %% function values and z1 < z2 for at least one objective function
  %% value.
  %%
  %% to find the pareto set points requires a double loop. this is
  %% therefore O(n^2) which is not great but, for the application
  %% intended when the code was written, function evaluations were so
  %% expensive that an O(n^2) computation here is not a big deal,
  %% especially for the size of n expected.
  pareto.x = [];
  pareto.z = [];
  for i = 1:n                         % for each design point
    %% is this point dominated?
    dominated = 0;                  % not so far
    for j = 1:n                     % compare it with every other
      if i ~= j                   % don't compare with oneself...
        if all (z(:,j) <= z(:,i)) && any (z(:,j) < z(:,i))
          dominated = 1;      % yup, this point is no good!
          break;              % leave the inner loop
        end
      end
    end
    if ~ dominated
      pareto.x = [pareto.x x(:,i)];
      pareto.z = [pareto.z z(:,i)];
    end
  end
  debugprint(1,'Pareto points and values: ', [pareto.x', pareto.z'])

  %% we need to sort the pareto set to have points in order, with respect
  %% to one of the objective function values. This only works for
  %% m=2!
  [xx indices] = sort (pareto.z (1, :));
  pareto.x = pareto.x (:, indices);
  pareto.z = pareto.z (:, indices);
  debugprint(1,'Pareto, now sorted: ', [pareto.x', pareto.z'])



  %% also note the range of values for each of the objective functions so
  %% that we can scale the fitness so that no single dimension
  %% dominates.
  zmin = min (z, [], 2);
  zmax = max (z, [], 2);
  %% make sure we cater for a degenerate (1 point) Pareto set
  if zmax <= zmin
    zmax = 1.0001 * zmin;
  end

  %% given a pareto set, we can define a distance for any point not in
  %% this set. this distance can be the fitness of a point. we don't
  %% need to scale here as we don't know how the fitness is going to
  %% be used. it may not need scaling (e.g. for a tournament
  %% selection procedure).
  np = size(pareto.x,2);
  fitness = zeros (1,n);              % fitness array to be returned
  linedistance = 0;                   % do fitness relative to nearest pareto point instead of to line segment
  for i = 1:n
    if 2 == m && np > 1 && linedistance
      %% the more intelligent approach to calculating a fitness is to
      %% determine the distance between the point and the most
      %% logical line segment from the set of line segments
      %% defined by the pareto set. which is the most logical
      %% line segment? one definition could be the line segment
      %% which intersects the line perpendicular to it that goes
      %% through the point whose fitness we wish to
      %% calculate. For this, we have to assume that the pareto
      %% set has been sorted to minimise the number of line
      %% segments to consider.
      fitness (i) = 1e10;
      for j = 2:np
        %% see if perpendicular line intersects the line segment defined by z_j
        %% and z_{j-1}.
        dy = pareto.z (2, j) - pareto.z (2, j-1);
        dx = pareto.z (1, j) - pareto.z (1, j-1);
        %% perpendicular line will have the inverse slope, both multiplicative
        %% and additive inverses in fact! we do need to be
        %% careful about extreme conditions (0 slope to
        %% infinity and vice versa)
        if abs (dy) < 1e-4 && abs (dx) < 1e-4
          %% 0 length line segment (means two identical points in the pareto set
          %% which we really should be avoiding but haven't
          %% managed to yet... for some reason)
          continue
        end
        pdy = -dx;
        pdx = dy;

      end
    else
      %% for each point i, simply find the distance to the nearest point in
      %% the pareto set.
      fitness (i) = 1e10;
      for j = 1:np
        d = norm ( (z (:,i) - pareto.z (:,j)) ./ (zmax-zmin), 2);
        if d < fitness (i)
          fitness (i) = d;
        end
      end
      %% check each point relative to the end points of the pareto front
      %% approximated by the set we have. we pretend that there
      %% are two extra points in the pareto set, one at each end,
      %% that extend the front vertically and horizontally to
      %% infinity with infinite and 0 slopes respectively.  this
      %% extra check is only used for when there are 2 objective
      %% functions.
      if 2 == m && np>0
        for j = 1:2
          k = 1;
          if 2 == j
            k = np;
          end
          d = (z (j,i) - pareto.z (j,k)) / (zmax (j) - zmin (j));
          if d > 0 && d < fitness (i)
            %% fprintf (stderr, 'Fitness based on extension of dimension: i=%d j=%d\r', i, j);
            fitness (i) = d;
          end
        end
      end
    end
  end
  debugprint (1, 'Population with fitness: ', [x', z', fitness'])
